import os
from PyQt5 import uic
from PyQt5.QtCore import QSettings
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QWidget, QWizard, QWizardPage, \
    QFileDialog, QMessageBox, QTableWidgetItem, QHeaderView

from qgis.core import QgsFeatureRequest, QgsCoordinateTransform, QgsCoordinateReferenceSystem, \
    QgsGeometry, QgsPoint, QgsMessageLog, QgsVectorLayer, \
    QgsPalLayerSettings, QgsVectorLayerSimpleLabeling
from qgis.gui import QgsMapCanvas

from .db import get_connection
from .paggf_utils import PAGGFUtils
from .paggf_layers import LoadedLayers
from .maptool_select import MapToolSelectRectangle


PAGE_1_IDC_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_idc_location.ui'))

PAGE_1_BM_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_bm_location.ui'))

PAGE_2_MANUAL_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_manual.ui'))

PAGE_2_FILE_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_file.ui'))

PAGE_3_VERTICES_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_selection.ui'))

class PageLocationIDC(QWizardPage, PAGE_1_IDC_W):
    def __init__(self, parent=None):
        super(PageLocationIDC, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Localisation de la parcelle")
        self.setSubTitle("Veuillez renseigner la localisation de la parcelle à importer")

    def initializePage(self):
        self.layers = LoadedLayers()
        self.utils = PAGGFUtils(self.layers)

        self.provinces_idx = {}
        self.municipalities_idx = {}
        self.hills_idx = {}
        self.sub_hills_idx = {}

        # populate combos
        province_layer = self.layers.get("ref", "provinces")
        if province_layer:
            idx=0
            req = QgsFeatureRequest().setSubsetOfAttributes([province_layer.fields().indexFromName(x) for x in ["id","name"]])
            for f in province_layer.getFeatures(req):
                self.province_field.addItem(
                    "{} - {}".format(f['id'], f['name']), f['id'])
                self.provinces_idx[f['id']]=idx
                idx+=1
        self.province_field.activated[int].connect(
            self.when_province_selection_changed)
        self.municipality_field.activated[int].connect(
            self.when_municipality_selection_changed)
        self.hill_field.activated[int].connect(
            self.when_hill_selection_changed)

        # select the current municipality, if any:
        current_municipality = self.utils.get_current_municipality()
        if current_municipality:
            id = current_municipality['id']
            self.province_field.setCurrentIndex(self.provinces_idx[id[0:2]])
            self.when_province_selection_changed(self.provinces_idx[id[0:2]])
            self.municipality_field.setCurrentIndex(self.municipalities_idx[id[0:4]])
            self.when_municipality_selection_changed(self.municipalities_idx[id[0:4]])
            self.province_field.setEnabled(False)
            self.municipality_field.setEnabled(False)
        else:
            self.province_field.setCurrentIndex(0)
            self.municipality_field.setCurrentIndex(0)
            self.hill_field.setCurrentIndex(0)
            self.subhill_field.setCurrentIndex(0)
            self.when_province_selection_changed(0)

    def when_province_selection_changed(self, index):
        self.municipalities_idx = {}
        mun_layer = self.layers.get("ref", "municipalities")
        self.municipality_field.clear()
        if mun_layer:
            idx = 0
            req = QgsFeatureRequest().setSubsetOfAttributes([mun_layer.fields().indexFromName(x) for x in ["id", "name"]])
            for f in mun_layer.getFeatures(req):
                if f['id'].startswith(self.province_field.currentData()):
                    self.municipality_field.addItem(
                        "{} - {}".format(f['id'], f['name']), f['id'])
                    self.municipalities_idx[f['id']] = idx
                    idx += 1
        self.when_municipality_selection_changed(0)

    def when_municipality_selection_changed(self, index):
        self.hills_idx = {}
        hill_layer = self.layers.get("ref", "hill_districts")
        self.hill_field.clear()
        if hill_layer:
            idx = 0
            req = QgsFeatureRequest().setSubsetOfAttributes([hill_layer.fields().indexFromName(x) for x in ["id", "name"]])
            for f in hill_layer.getFeatures(req):
                if f['id'].startswith(self.municipality_field.currentData()):
                    self.hill_field.addItem(
                        "{} - {}".format(f['id'], f['name']), f['id'])
                    self.hills_idx[f['id']] = idx
                    idx += 1
        self.when_hill_selection_changed(0)

    def when_hill_selection_changed(self, index):
        """
        Called when the user change the selection of the hill (the selected
        hill in the combo-box hill_field has changed)

        Params
        ------
        int index : new index in the combobox of the selected hill
        """
        self.sub_hills_idx = {}
        hill_id = self.hill_field.itemData(index)
        self.subhill_field.clear()
        sub_hill_layer = self.layers.get("ref", "sub_hill_avenues")
        if sub_hill_layer:
            idx = 0
            req = QgsFeatureRequest().setSubsetOfAttributes([sub_hill_layer.fields().indexFromName(x) for x in ["id", "name"]])
            for f in sub_hill_layer.getFeatures(req):
                if f['id'].startswith(self.hill_field.currentData()):
                    self.subhill_field.addItem(
                        "{} - {}".format(f['id'], f['name']), f['id'])
                    self.sub_hills_idx[f['id']] = idx
                    idx += 1

        # mise à jour numéro parcelle
        conn = get_connection()
        cur = conn.cursor()
        cur.execute("select gis.next_parcel_num('{}')".format(hill_id))
        parcel_number = cur.fetchone()[0]
        self.parcel_number_field.setText(str(parcel_number))

    def subhill(self):
        return self.subhill_field.currentData()

    def parcel_number(self):
        return int(self.parcel_number_field.text())
        
class PageLocationBM(QWizardPage, PAGE_1_BM_W):
    def __init__(self, parent=None):
        super(PageLocationBM, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Identification de la parcelle")
        self.setSubTitle("Veuillez renseigner l'identification de la parcelle")

    def initializePage(self):
        conn = get_connection()
        cur = conn.cursor()
        # TODO: METTRE DANS UN FICHIER REQUETES SQL

        cur.execute("\
select id, commission_reference, idc_parcel_code FROM gis.idc_parcels")
        # TODO utiliser gis.valid_idc_parcels

        self.commission_reference_choice.addItem('---', None)
        self.idc_parcel_code_choice.addItem('---', None)

        for cur_row in cur:
            self.commission_reference_choice.addItem(cur_row[1], cur_row[0])

            if cur_row[2]:
                self.idc_parcel_code_choice.addItem(cur_row[2], cur_row[0])
            else:
                self.idc_parcel_code_choice.addItem('sans référence', cur_row[0])

        conn.close()

    def id_parcel(self):
        if self.radio_button_idc_parcel.isChecked():
            id_parcel = self.idc_parcel_code_choice.currentData()
        else:
            id_parcel = self.commission_reference_choice.currentData()
        return id_parcel

class PageManualEdit(QWizardPage, PAGE_2_MANUAL_W):
    def __init__(self, parent=None):
        super(PageManualEdit, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Saisie des sommets")
        self.setSubTitle("Veuillez saisir les coordonnées des sommets")

    def initializePage(self):
        self.tableWidget.setRowCount(1)
        self.tableWidget.setColumnCount(3)

        self.tableWidget.setHorizontalHeaderItem(0, QTableWidgetItem('X'))
        self.tableWidget.setHorizontalHeaderItem(1, QTableWidgetItem('Y'))
        self.tableWidget.setHorizontalHeaderItem(2, QTableWidgetItem('Name'))

        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch)

        self.tableWidget.setItem(0, 0, QTableWidgetItem())
        self.tableWidget.setItem(0, 1, QTableWidgetItem())
        self.tableWidget.setItem(0, 2, QTableWidgetItem())

        self.pushButton.clicked.connect(self.add_row)
        self.tableWidget.cellChanged.connect(self.check_cell)

    def check_cell(self, row, column):
        error = False
        for row in range(self.tableWidget.rowCount()):
            for column in range(2):
                item = self.tableWidget.item(row, column)
                if item:  # si possede widget
                    if item.text() and not item.text().isdigit():
                        item.setBackground(QBrush(QColor(200, 0, 0)))
                        error = True
                    else:
                        item.setBackground(QBrush())

    def add_row(self):
        row_count = self.tableWidget.rowCount()

        self.tableWidget.setRowCount(self.tableWidget.rowCount() + 1)
        self.tableWidget.setItem(row_count, 0, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 1, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 2, QTableWidgetItem())

    def get_features(self):
        for row in range(self.tableWidget.rowCount()):
            x = self.tableWidget.item(row, 0).text()
            y = self.tableWidget.item(row, 1).text()
            if x and y:  # avoid empty cell
                name = self.tableWidget.item(row, 2).text()
                geom = QgsGeometry(QgsPoint(float(x), float(y)))

                yield ({'name':name}, geom)

    def get_name_field(self):
        return 'name'
    
class PageFile(QWizardPage, PAGE_2_FILE_W):
    def __init__(self, parent=None):
        super(PageFile, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Fichier à importer")
        self.setSubTitle("Veuillez sélectionner le fichier de sommets à importer")

        self.browse_btn.clicked.connect(self.on_browse)
        # the file path is mandatory
        self.registerField("file_path*", self.file_edit)
        
        self._layer = None

    def on_browse(self):
        s = QSettings("PAGGF", "PAGGF")
        last_shp_path = s.value("lastSHPParcelPath", ".")

        if not os.path.exists(last_shp_path):
            last_shp_path = os.getcwd()
        shp_path, _ = QFileDialog.getOpenFileName(
            self, "Sélectionnez le fichier de sommets",
            last_shp_path,
            "*.{}".format(self.file_type_combo.currentText().lower()))

        s.setValue("lastSHPParcelPath", shp_path)
        QgsMessageLog.logMessage(shp_path, 'PAGGF', QgsMessageLog.ALL)

        self.file_edit.setText(shp_path)

    def file_path(self):
        return self.file_edit.text()

    def layer(self):
        return self._layer

    def validatePage(self):
        layer_file_path = self.file_path()
        layer_file_path_ext = os.path.splitext(layer_file_path)[1]

        provider_lib = "ogr"  # Default provider lib  for shp
        path = self.file_path()  # Default uri for shp
        if layer_file_path_ext.lower() == '.csv':
            provider_lib = "delimitedtext"
            path = "\
file://{}?delimiter={}&crs=epsg:32735&xField={}&yField={}".format(
                self.shp_path(),
                ";", "X", "Y")
        elif layer_file_path_ext.lower() == '.gpx':
            provider_lib = "gpx"
            path = "{}?type=waypoint".format(self.file_path())

        self._layer = QgsVectorLayer(
            path, "data_import", provider_lib)
        if not self._layer.isValid():
            QMessageBox.warning(None, "Problème de chargement", "Problème lors du chargement de la couche")
            return False

        return True

class PageVertexSelection(QWizardPage, PAGE_3_VERTICES_W):
    def __init__(self, parent=None):
        super(PageVertexSelection, self).__init__(parent)
        self.setupUi(self)
        self._parent = parent

        self.setTitle("Sélection des sommets et attributs")
        self.setSubTitle("Veuillez sélectionner les sommets à importer et renseigner le nom des champs")

    def initializePage(self):
        self.canvas.setLayers([self._parent.layer()])
        extent = self._parent.layer().extent()
        # add 20% around the extent
        extent.scale(1.2)
        self.canvas.setExtent(extent)

        # by default, select all the features
        self._parent.layer().selectAll()
        
        mt = MapToolSelectRectangle(self.canvas, self._parent.layer())
        self.canvas.setMapTool(mt)

        # re-evaluate isComplete on each selection change
        self._parent.layer().selectionChanged.connect(self.completeChanged)

        # register fields
        self.registerField("name_field", self.name_field, "currentText")
        
        # populate combos to select fields
        for field in self._parent.layer().fields():
            if not field.isNumeric():
                self.name_field.addItem(field.name())

        self.name_field.currentTextChanged.connect(self.on_name_field_changed)
        self.on_name_field_changed()

    def on_name_field_changed(self):
        # update the labeling options
        label_settings = QgsPalLayerSettings()
        label_settings.fieldName = self.name_field.currentText()
        self._parent.layer().setLabeling(QgsVectorLayerSimpleLabeling(label_settings))
        self.canvas.refreshAllLayers()

    def get_selected(self):
        return self._parent.layer().getSelectedFeatures()

    # iterator that returns (feature, transformed_geometry)
    def get_features(self):
        layer = self._parent.layer()
        srid_layer = layer.crs().postgisSrid()

        transf = None
        if srid_layer != 32735:
            transf = QgsCoordinateTransform(
                QgsCoordinateReferenceSystem(srid_layer),
                QgsCoordinateReferenceSystem(32735))

        for f in layer.getSelectedFeatures():
            if transf:
                geom = QgsGeometry(QgsPoint(transf.transform(f.geometry().asPoint())))
            else:
                geom = f.geometry()

            yield (f, geom)

    def isComplete(self):
        return self._parent.layer().selectedFeatureCount() > 0

    def get_name_field(self):
        return self.name_field.currentText()

class LoadIDCParcelWizard(QWizard):
    def __init__(self, from_file, parent=None):
        super(LoadIDCParcelWizard, self).__init__(parent)
        self.setWindowTitle("Assistant pour l'import de parcelle CID")
        self.location_page = PageLocationIDC(self)
        self.addPage(self.location_page)
        if from_file:
            self.file_page = PageFile(self)
            self.addPage(self.file_page)
            self.selection_page = PageVertexSelection(self)
            self.addPage(self.selection_page)
        else:
            self.selection_page = PageManualEdit(self)
            self.addPage(self.selection_page)

    def layer(self):
        return self.file_page.layer()

    def get_features(self):
        return self.selection_page.get_features()

    def subhill(self):
        return self.location_page.subhill()
    def parcel_number(self):
        return self.location_page.parcel_number()
    def name_field(self):
        return self.selection_page.get_name_field()

class LoadBMParcelWizard(QWizard):
    def __init__(self, from_file, parent=None):
        super(LoadBMParcelWizard, self).__init__(parent)
        self.setWindowTitle("Assistant pour l'import de parcelle BM")
        self.location_page = PageLocationBM(self)
        self.addPage(self.location_page)
        if from_file:
            self.file_page = PageFile(self)
            self.addPage(self.file_page)
            self.selection_page = PageVertexSelection(self)
            self.addPage(self.selection_page)
        else:
            self.selection_page = PageManualEdit(self)
            self.addPage(self.selection_page)

    def layer(self):
        return self.file_page.layer()

    def get_features(self):
        return self.selection_page.get_features()

    def id_parcel(self):
        return self.location_page.id_parcel()
    def name_field(self):
        return self.selection_page.get_name_field()

